/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
   Copyright (C) 2000 CodeFactory AB
   Copyright (C) 2000 Jonas Borgström <jonas@codefactory.se>
   Copyright (C) 2000 Anders Carlsson <andersca@codefactory.se>
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
   
   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include <gtk/gtk.h>
#include <string.h>
#include <stdlib.h>
#include <libxml/uri.h>
#include <libxml/debugXML.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <libgen.h>
#include <dirent.h>

#include <libgtkhtml/graphics/htmlpainter.h>
#include <libgtkhtml/util/htmlstream.h>
#include <libgtkhtml/gtkhtmlcontext.h>

#include "browser-window.h"
#include "url-utils.h"
#include "glibcurl.h"
#include "history.h"

#define CONNECTION_CLEANER_PERIOD 1000 /* 1s */
#define CONNECTION_TIMEOUT 10 /* 5s since last receive */

#define APPNAME "GtkHtml II"
#define USERAGENT "Mozilla/5.0 (compatible; libgtkhtml)"

static void browser_window_init		(BrowserWindow		 *window);
static void browser_window_class_init	(BrowserWindowClass	 *klass);
static void link_clicked (HtmlDocument *doc, const gchar *url, gpointer data);
static void url_requested (HtmlDocument *doc, const gchar *uri, HtmlStream *stream, gpointer data);
static void browser_window_entry_activate (GtkWidget *widget, gpointer data);

static GtkWindowClass *parent_class = NULL;

// ============= stream management ============

G_LOCK_DEFINE_STATIC(connlist);

typedef struct {
	// persistent objects, no need to delete
	HtmlDocument *doc;
	HtmlStream *stream;
	BrowserWindow *window;
	
	// things we need to cleanup
	CURL *handle;
	URL *url;
	gchar *urlstr;
	
	// flags
	long content_length;
	long bytes_read;
	int conn_timeout;
	gboolean is_closed;
} StreamData;


static void
free_stream_data (StreamData *sdata)
{
	if (!sdata) return;
	if (!sdata->is_closed) {
		sdata->is_closed = TRUE;
		//g_print("Closing stream for %s\n", sdata->urlstr);	
		glibcurl_remove (sdata->handle);
		curl_easy_cleanup(sdata->handle);
		html_stream_close(sdata->stream);

		url_free (sdata->url);
		g_free (sdata->urlstr);
		sdata->url = sdata->urlstr = sdata->stream = sdata->handle = 0;
	}
}

static void add_connection (HtmlDocument *doc, StreamData *sdata) {
	GSList *connection_list;
	
	//g_print("add %s\n", sdata->urlstr);
	G_LOCK(connlist);
	connection_list = g_object_get_data (G_OBJECT (doc), "connection_list");
	connection_list = g_slist_prepend (connection_list, sdata);
	g_object_set_data (G_OBJECT (doc), "connection_list", connection_list);	
	G_UNLOCK(connlist);
}


void close_stale_connections (HtmlDocument *doc, gboolean force)
{	
	GSList *connection_list, *tmp;
	G_LOCK(connlist);
	tmp = connection_list = g_object_get_data (G_OBJECT (doc), "connection_list");
	G_UNLOCK(connlist);

	while (tmp) {
		StreamData *sdata = (StreamData *)tmp->data;
		gboolean close = FALSE;
		
		if (sdata->content_length>0 && 
		    sdata->content_length == sdata->bytes_read)
				close = TRUE;				

		if (sdata->conn_timeout > CONNECTION_TIMEOUT)
			close = TRUE;
		else
			sdata->conn_timeout++;

		tmp = tmp->next;

		if (close || force) {
			free_stream_data (sdata);
		}		
	}
}

void remove_closed_connections (HtmlDocument *doc)
{	
	G_LOCK(connlist);
	GSList *connection_list, *tmp;
	tmp = connection_list = g_object_get_data (G_OBJECT (doc), "connection_list");

	while (tmp) {
		StreamData *sdata = (StreamData *)tmp->data;
		tmp = tmp->next;
		if (sdata->is_closed) {
			connection_list = g_slist_remove (connection_list, sdata);
			free(sdata);
		}
	}
	g_object_set_data (G_OBJECT (doc), "connection_list", connection_list);
	G_UNLOCK(connlist);
}

static int on_idle_remove_stale_connections (gpointer data)
{
	// two-step close: first close the connections (short-time locking),
	// and then removal (long-time locking)
	close_stale_connections (((BrowserWindow *) data)->doc, FALSE);
	remove_closed_connections (((BrowserWindow *) data)->doc);
	return G_SOURCE_CONTINUE;
}


// ==== stream callbacks  (initiated by url_requested);

static void
stream_cancel (HtmlStream *stream, gpointer user_data, gpointer cancel_data)
{
	StreamData *sdata = (StreamData *)cancel_data;
	if (sdata) free_stream_data (sdata);
}

size_t stream_write_callback(char *buffer, size_t size, size_t nmemb, void *userdata)
{
	StreamData *sdata = (StreamData *) userdata;
	long bytes_read = size*nmemb;
	if (sdata->is_closed) return -1;
	
	if (!sdata->content_length) {
		double content_length;
		if (CURLE_OK == curl_easy_getinfo(sdata->handle, 
		                CURLINFO_CONTENT_LENGTH_DOWNLOAD, 
		                &content_length))
			sdata->content_length = (long)content_length;
		else
			sdata->content_length = -1;
		//g_print("%s - content-length: %ld\n", sdata->urlstr, sdata->content_length);			
	}
	
	sdata->bytes_read += bytes_read;	
	//g_print("%s - bytes read: %ld (%ld) \n", sdata->urlstr,bytes_read, sdata->bytes_read);
	if (bytes_read) {
		html_stream_write (sdata->stream, buffer, bytes_read);
		sdata->conn_timeout = 0; // reset timeout if we have new data
	}
	return bytes_read;
}

typedef struct {
	gchar *url;
	BrowserWindow *window;
} RedirectContext;

static int
on_idle_redirect (gpointer data) {
	RedirectContext *ctx = (RedirectContext *)data;
	link_clicked (ctx->window->doc, ctx->url, ctx->window);
	g_free(ctx->url);
	g_free(ctx);
	return FALSE;
}

size_t stream_header_callback(char *buffer, size_t size, size_t nmemb, void *userdata)
{
	StreamData *sdata = (StreamData *) userdata;
	long bytes_read = size*nmemb;
	if (sdata->is_closed) return -1;

	//g_print("bytes read: %ld\n",bytes_read);	
	if (bytes_read) {
		if (g_str_has_prefix(buffer, "Location: ")) {
			gchar* p = g_strstrip(g_strdup(buffer + 10)); // length of "Location: "
			g_print("Redirect to: %s\n", p);
			//url_requested (sdata->doc, p, sdata->stream, sdata->window);
			//g_free(p);			
			// we could the above, but then we can't display the new URL on our entry box
			// so let the request expire instead and do it on idle
			// calling link_clicked directly from here will cause segfaults
			RedirectContext *ctx = g_new0(RedirectContext, 1);
			ctx->window = sdata->window;
			ctx->url = p;
			g_idle_add (on_idle_redirect, ctx);
		}
	}
	return bytes_read;
}

static void 
load_file (HtmlDocument *doc, const gchar *name, HtmlStream *stream) 
{
	int fd;
	struct stat sb;	
	gchar *buffer;
	GSList *files=NULL, *dirs=NULL, *tmp;
	
	if (!stream) html_document_clear (doc);

    /* we do assume UTF-8 encoding */
    if ((fd=open(name,O_RDONLY))>=0) {
		fstat(fd, &sb);
		
		if (!S_ISDIR(sb.st_mode)) {
			// normal file, load it
			buffer = mmap(NULL, sb.st_size, PROT_READ, MAP_SHARED | MAP_NORESERVE, fd, 0);
			if (buffer!=(gchar *)-1) {
				if (stream) {
					html_stream_write (stream, buffer, sb.st_size);
				} else if (html_document_open_stream (doc, "text/html")) {
					html_document_write_stream (doc, buffer, sb.st_size);
					html_document_close_stream (doc);
				}
				munmap(buffer, sb.st_size);
			}
			close(fd);
		} else {
			// dir, show index
			DIR *dir = fdopendir(fd);
			struct dirent *entry;
			if (stream || html_document_open_stream (doc, "text/html")) {
				//header
				buffer = g_strdup_printf("<big><b>%s</b></big><hr><br><br>\n", name);
				if (stream) html_stream_write (stream, buffer, strlen(buffer));
				else html_document_write_stream (doc, buffer, strlen(buffer));
				
				//table start
				buffer = g_strdup_printf("<table><tr><td>Name</td></tr><td>Type</td>\n", name);
				if (stream) html_stream_write (stream, buffer, strlen(buffer));
				else html_document_write_stream (doc, buffer, strlen(buffer));
				
				// entries - load and sort, then display
				while ((entry = readdir(dir)) != NULL)
				{
					stat(entry->d_name, &sb);
					if (S_ISDIR(sb.st_mode)) {
						if (entry->d_name[0]=='.' && !entry->d_name[1]) continue;
						dirs = g_slist_prepend (dirs, g_strdup(entry->d_name));
					} else {
						files = g_slist_prepend (files, g_strdup(entry->d_name));
					}
				}
				dirs = g_slist_sort(dirs, (GCompareFunc) g_strcasecmp);
				files = g_slist_sort(files, (GCompareFunc) strcasecmp);
				
				tmp = dirs;
				while (tmp) {
					buffer = g_strdup_printf("<tr><td><a href=\"%s\">%s</a><td><td>&lt;DIR&gt;</td></tr>\n", 
					tmp->data, tmp->data);
					if (stream) html_stream_write (stream, buffer, strlen(buffer));
					else html_document_write_stream (doc, buffer, strlen(buffer));
					g_free(buffer);					
					tmp = tmp->next;
				}
				tmp = files;
				while (tmp) {
					buffer = g_strdup_printf("<tr><td><a href=\"%s\">%s</a><td><td>File</td></tr>\n", 
					tmp->data, tmp->data);
					if (stream) html_stream_write (stream, buffer, strlen(buffer));
					else html_document_write_stream (doc, buffer, strlen(buffer));
					g_free(buffer);					
					tmp = tmp->next;
				}
				
				g_slist_free_full (dirs, g_free);
				g_slist_free_full (files, g_free);

				//table end
				buffer = g_strdup_printf("</table>", name);
				if (stream) html_stream_write (stream, buffer, strlen(buffer));
				else {
					html_document_write_stream (doc, buffer, strlen(buffer));				
					html_document_close_stream (doc);
				}
			}
			closedir(dir);
		}
	} else {
		g_print ("Could not open file %s\n", name);
	}
}


// ==== URI helpers
// return TRUE if same as base, FALSE OTHERWISE:
static gboolean
set_base (BrowserWindow *window, const gchar *url)
{
	//g_print("proto: %s, host: %s, user: %s, pass: %s, path: %s, query: %s, anchor: %s\n", 
	//url_get_proto(url),url_get_host(url),url_get_user(url),url_get_pass(url),url_get_path(url),url_get_query(url),url_get_anchor(url));
	//URL *u = str_to_url(url);
	//g_printf("-->%s<--\n", url_to_str(u));
	//return FALSE;
	gboolean equal = FALSE;
	
	if (window->baseURI) {
		URL* new_base = url_set_relative (window->baseURI, url);
		
		gchar *old = url_to_str(window->baseURI);
		gchar *new = url_to_str(new_base);
		
		if (strchr(old,'#')) *(strchr(old,'#'))=0;
		if (strchr(new,'#')) *(strchr(new,'#'))=0;
		equal = (strcmp(old,new)==0);
		
		url_free (window->baseURI);
		window->baseURI = new_base;		
		g_free(old);
		g_free(new);
		
	} else {
		window->baseURI = str_to_url (url);
	}

	gchar *new_uri = url_to_str(window->baseURI);	
	if (!equal) g_print("baseURI: %s\n", new_uri);	
	gtk_entry_set_text (GTK_ENTRY (window->entry), new_uri);
	g_free(new_uri);		
	return equal;		
}


// === htmldoc event handlers

typedef struct {
	BrowserWindow *window;
	gchar *action;
	gchar *method;
	gchar *encoding;
} SubmitContext;

static int
on_idle_submit (gpointer data)
{
	SubmitContext *ctx = (SubmitContext *)data;

	g_print ("action = '%s', method = '%s', encoding = '%s'\n", 
		 ctx->action, ctx->method, ctx->encoding);

	if (ctx->method == NULL || strcasecmp (ctx->method, "get") == 0) {
		gchar *url;

		url = g_strdup_printf ("%s?%s", ctx->action, ctx->encoding);
		link_clicked (NULL, url, ctx->window);
		g_free (url);
	}
	g_free (ctx);
	return 0;
}

static void
on_submit (HtmlDocument *document, const gchar *action, const gchar *method, 
	   const gchar *encoding, gpointer data)
{
	SubmitContext *ctx = g_new0 (SubmitContext, 1);

	if (action)
		ctx->action = g_strdup (action);
	if (method)
		ctx->method = g_strdup (method);
	if (action)
		ctx->encoding = g_strdup (encoding);
	ctx->window = data;

	/* Because the link_clicked method will clear the document and
	 * start loading a new one, we can't call it directly, because
	 * gtkhtml2 will crash if the document becomes deleted before
	 * this signal handler finish */
	g_idle_add (on_idle_submit, ctx);
}

static void
url_requested (HtmlDocument *doc, const gchar *uri, HtmlStream *stream, gpointer data)
{
	BrowserWindow *window = (BrowserWindow *)data;
	StreamData *sdata;
	gchar *urlstr;
	URL *url;

	g_assert (HTML_IS_DOCUMENT(doc));
	g_assert (stream != NULL);

	url = url_set_relative (window->baseURI, uri);
	urlstr = url_to_str(url);
	
	if (!window->allow_online || !url->proto || strcmp(url->proto,"file")==0) {
		html_stream_set_cancel_func (stream, stream_cancel, NULL);
		load_file (doc, url->path, stream);
		html_stream_close (stream);
		url_free (url);
		g_free(urlstr);
		return;
	}

	sdata = g_new0 (StreamData, 1);
	sdata->window = window;
	sdata->doc = doc;
	sdata->stream = stream;
	sdata->url = url;
	sdata->urlstr = urlstr;
	//g_print ("uri requested: %s\n", uri);
	//g_print ("URL requested: %s\n", sdata->urlstr);

	CURL *handle = curl_easy_init();
	sdata->handle = handle;
	curl_easy_setopt (handle, CURLOPT_URL, sdata->urlstr);
	curl_easy_setopt (handle, CURLOPT_WRITEFUNCTION, stream_write_callback);
	curl_easy_setopt (handle, CURLOPT_WRITEDATA, sdata);
	curl_easy_setopt (handle, CURLOPT_USERAGENT, USERAGENT);
	//curl_easy_setopt (handle, CURLOPT_FOLLOWLOCATION, 1);
	curl_easy_setopt (handle, CURLOPT_HEADERFUNCTION, stream_header_callback);
	curl_easy_setopt (handle, CURLOPT_HEADERDATA, sdata);
	curl_easy_setopt (handle, CURLOPT_NOSIGNAL, 1L);
	//curl_easy_setopt (handle, CURLOPT_TIMEOUT, 20L); // 20s timeout
	if (window->allow_insecure) {
		curl_easy_setopt (handle, CURLOPT_SSL_VERIFYPEER, 0L);
		curl_easy_setopt (handle, CURLOPT_SSL_VERIFYHOST, 0L);	
	}
	if (window->debug_connection) {
		curl_easy_setopt (handle, CURLOPT_VERBOSE, 1L);
	}
	
	glibcurl_add (handle);

	html_stream_set_cancel_func (stream, stream_cancel, sdata);
	add_connection (doc, sdata);
}

static void
title_changed (HtmlDocument *doc, const gchar *new_title, gpointer data)
{
	BrowserWindow *window = (BrowserWindow *) data;
	gchar *str = g_strconcat (window->title_prefix, ": ", new_title, NULL);
	gtk_window_set_title (GTK_WINDOW (window), str);
	g_free (str);
}

static void
link_clicked (HtmlDocument *doc, const gchar *url, gpointer data)
{
	BrowserWindow *window = (BrowserWindow *)data;
	const gchar *anchor;
	gchar *str_url;
	gboolean equal;

	g_print ("signal \"link_clicked\" url = \"%s\"\n", url);

	if(strcmp(url,"")==0) return;
	
	if(!window->allow_online) {
		URL *u = url_new();
		u = str_to_url(url);
		if(u->proto != NULL && strcmp(u->proto,"file")!=0) {
			// Offer opening the link in an external browser
			GtkResponseType result;
			GtkWidget *dialog = gtk_message_dialog_new(GTK_WINDOW(window),
				GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO,
				"Only local files are supported in off-line mode!\n"
				"Click Yes to open this link in your default browser:\n"
				"%s", url);
			gtk_window_set_title(GTK_WINDOW(dialog), "Request");
			result = gtk_dialog_run(GTK_DIALOG(dialog));
	
			if(result == GTK_RESPONSE_YES) {
				gchar buf[BUFSIZ];
				snprintf(buf, sizeof(buf), "defaultbrowser %s &", url);
				system(buf);
			}
			gtk_widget_destroy(dialog);
	
			url_free(u);
			return;
		}
		url_free(u);
	}

	equal = set_base (window, url);
	str_url = url_to_str(window->baseURI);
	anchor = window->baseURI->anchor;

	if (equal && !window->reload && anchor != NULL) {
		/* Same document, just jump to the anchor */
		html_view_jump_to_anchor (window->view, anchor);
	}
	else {		
		close_stale_connections (window->doc, TRUE);

		/* Now load the specified filename */		
		html_document_clear (window->doc);
		html_document_open_stream (window->doc, "text/html");
		gtk_adjustment_set_value (gtk_layout_get_vadjustment (GTK_LAYOUT (window->view)), 0);

		url_requested (window->doc, str_url, window->doc->current_stream, data);
		
		if (window->baseURI) {
			anchor = window->baseURI->anchor;
			if (anchor != NULL)
				html_view_jump_to_anchor (window->view, anchor);
		}
	}	
	if (!window->nohistory) {
		history_push (&window->back_history, str_url);
		history_clear (&window->forward_history);
	}	

	window->reload = FALSE;			
	g_free (str_url);
	
	gtk_widget_queue_resize (GTK_WIDGET (data));
	gtk_label_set_text (GTK_LABEL (GTK_STATUSBAR (window->status_bar)->label), "");
}

// ========= htmlview event handlers

static gboolean
request_object (HtmlView *view, GtkWidget *widget, gpointer user_data)
{
	// object handler - we don't support for now.
/*
	GtkWidget *sel;

	sel = gtk_color_selection_new ();
	gtk_widget_show (sel);

	gtk_container_add (GTK_CONTAINER (widget), sel);
*/
	return TRUE;
}

static void
on_url (HtmlView *view, const char *url, gpointer user_data)
{
	BrowserWindow *window = BROWSER_WINDOW (user_data);

	gtk_label_set_text (GTK_LABEL (GTK_STATUSBAR (window->status_bar)->label), 
			    url);
}

// ==== browser window internal functions

static void
browser_window_set_view (BrowserWindow *window, HtmlView *view)
{
	GtkWidget *scr;

	window->view = view;

	g_signal_connect (G_OBJECT (view), "request_object",
			  G_CALLBACK (request_object), NULL);

	g_signal_connect (G_OBJECT (view), "on_url",
			  G_CALLBACK (on_url), window);

	scr = gtk_scrolled_window_new (gtk_layout_get_hadjustment (GTK_LAYOUT (view)),
				       gtk_layout_get_vadjustment (GTK_LAYOUT (view)));
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scr), GTK_POLICY_AUTOMATIC,
					GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER (scr), GTK_WIDGET (view));
	gtk_box_pack_start (GTK_BOX (window->vbox), scr, TRUE, TRUE, 0);
	gtk_widget_show_all (scr);
}


// ==== browser window event handlers

static void
browser_window_exit (gpointer data, guint action, GtkWidget *widget)
{
	gtk_main_quit ();
}

static void
browser_window_new_view (gpointer data, guint action, GtkWidget *widget)
{
	BrowserWindow *window  = BROWSER_WINDOW (data);
	GtkWidget *new_window;

	/* Create a new view */
	new_window = browser_window_new (window->doc);
	gchar *cur_url = url_to_str (window->baseURI);
	set_base (BROWSER_WINDOW(new_window), cur_url);
	g_free (cur_url);

	gtk_widget_show (new_window);
	
}

static gint
browser_window_reload (gpointer data, guint action, GtkWidget *widget)
{
	BrowserWindow *window  = BROWSER_WINDOW (data);
	window->reload = window->nohistory = TRUE;
	browser_window_entry_activate (window->entry, data);
	window->nohistory = FALSE;
	return FALSE;
}

static gint
browser_window_reload_delayed (gpointer data, GdkEvent *event, GtkWidget *widget)
{
	static guint id = 0;
	static guint old_width = 0;
	guint new_width = event->configure.width;

	if ( old_width != new_width ) {
		if (id) g_source_remove (id);

		id = g_timeout_add (100, browser_window_reload, data);
		old_width = new_width;
	}

	return FALSE;
}

static GtkWidget *file_open_dialog;
static void
browser_window_open_file (gpointer data, guint action, GtkWidget *widget)
{
	BrowserWindow *window  = BROWSER_WINDOW (data);
	
	if (!file_open_dialog) {
		file_open_dialog = gtk_file_chooser_dialog_new ("Open File",
						  GTK_WINDOW(window),
						  GTK_FILE_CHOOSER_ACTION_OPEN,
						  GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
						  GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
						  NULL);
	}
	if (gtk_dialog_run (GTK_DIALOG (file_open_dialog)) == GTK_RESPONSE_ACCEPT)
	{
		// loaded local file is not stored in history - on purpose.
		// to store in history, use browser_window_load_url instead.
		gchar *filename;
		filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (file_open_dialog));
		browser_window_load_file (GTK_WIDGET(window), filename);
	}
	gtk_widget_hide (file_open_dialog);
}

static void
browser_window_help_about (gpointer data, guint action, GtkWidget *widget)
{
	gtk_show_about_dialog(GTK_WINDOW(data), 
		"program-name", APPNAME " browser",
		"copyright", "© libgtkhtml2 authors, 2000\n© James B, 2015",
		"license", "GNU GPL Version 2 or later.", 
		"title", APPNAME,
		NULL
	);
}

static void
browser_window_close (gpointer data, guint action, GtkWidget *widget)
{
	BrowserWindow *window  = BROWSER_WINDOW (data);
	gtk_widget_destroy(GTK_WIDGET(window));
}

static void
browser_window_entry_activate (GtkWidget *widget, gpointer data)
{
	const gchar *str = gtk_entry_get_text (GTK_ENTRY (widget));
	gchar *url;

	BrowserWindow *window  = BROWSER_WINDOW (data);

	/*
	 * If no protocol is specified, then use file (offline) or http (online)
	 */
	if (strchr (str, ':'))
		url = g_strdup (str);
	else if (window->allow_online)
		url = g_strdup_printf ("http://%s", str);
	else
		url = g_strdup_printf ("file://%s", str);

	link_clicked (NULL, url, data);

	g_free (url);
}

static void
zoom_changed (GtkSpinButton *spin_button, gpointer user_data)
{
	html_view_set_magnification (BROWSER_WINDOW (user_data)->view,
				     gtk_spin_button_get_value (spin_button) / 100.0);
}

static void 
browser_window_back_clicked (GtkWidget *widget, gpointer data) {
	BrowserWindow *window  = BROWSER_WINDOW (data);
	gchar *current = history_pop (&window->back_history);
	gchar *prev = history_peek (&window->back_history);
	if (prev) {
		window->nohistory = TRUE;
		//g_print("back to %s\n",prev);
		history_push (&window->forward_history, current);
		link_clicked (window->doc, prev, data);
		g_free(prev);
		g_free(current);
		window->nohistory = FALSE;
	} else if (current) {
		// can't go back, put popped history back.
		history_push (&window->back_history, current);
		//g_print("can't go back\n");
		g_free(current);		
	}
}

static void 
browser_window_forward_clicked (GtkWidget *widget, gpointer data) {
	BrowserWindow *window  = BROWSER_WINDOW (data);
	gchar *next = history_pop (&window->forward_history);
	if (next) {
		window->nohistory = TRUE;
		//g_print("forward to %s\n",next);
		history_push (&window->back_history, next);
		link_clicked (window->doc, next, data);
		g_free(next);
		window->nohistory = FALSE;
	} else {
		//g_print("can't go forward\n");
	}
}

static void 
browser_window_load_home (gpointer data, guint action, GtkWidget *widget) {
	BrowserWindow *window  = BROWSER_WINDOW (data);
	link_clicked (NULL, window->home_url, data);
}

static void 
browser_window_stop_loading (gpointer data, guint action, GtkWidget *widget) {
	BrowserWindow *window  = BROWSER_WINDOW (data);
	close_stale_connections (window->doc, TRUE);
}

static void
browser_window_set_online_toggle(gpointer data, guint action, GtkWidget *widget) {
	BrowserWindow *window = BROWSER_WINDOW (data);

	if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(widget)) != window->allow_online) {
		window->allow_online = !window->allow_online;
	}
}

static GtkItemFactoryEntry menu_items[] = {
	{ "/_File",          NULL,         NULL, 0, "<Branch>" },
	{ "/File/_Open",     "<Control>O", (GtkItemFactoryCallback1)browser_window_open_file, 0, NULL },	
	{ "/File/_New View", "<Control>N", (GtkItemFactoryCallback1)browser_window_new_view, 0, NULL },
	{ "/File/_Home",     NULL,         (GtkItemFactoryCallback1)browser_window_load_home, 0, NULL },
	{ "/File/sep1",      NULL,         NULL, 0, "<Separator>" },
	{ "/File/_Stop",     NULL,         (GtkItemFactoryCallback1)browser_window_stop_loading, 0, NULL },	
	{ "/File/_Reload",   "F5",         (GtkItemFactoryCallback1)browser_window_reload, 0, NULL },
	{ "/File/_Close",    "<Control>W", (GtkItemFactoryCallback1)browser_window_close, 0, NULL },
	{ "/File/sep1",      NULL,         NULL, 0, "<Separator>" },
	{ "/File/On_line",   "<Control>L", (GtkItemFactoryCallback1)browser_window_set_online_toggle, 0, "<CheckItem>" },
	{ "/File/sep1",      NULL,         NULL, 0, "<Separator>" },
	{ "/File/E_xit",     "<Control>Q", (GtkItemFactoryCallback1)browser_window_exit, 0, NULL },
	{ "/_Help",           NULL,        NULL, 0, "<Branch>" },
	{ "/Help/_About",     NULL,        (GtkItemFactoryCallback1)browser_window_help_about, 0, NULL },
};

static void
browser_window_init (BrowserWindow *window)
{
	GtkWidget *hbox, *frame, *spinner;
	GtkWidget *back_button, *forward_button, *refresh_button;
	GtkWidget *home_button, *stop_button;
	GtkAccelGroup *accels;

	window->doc = NULL;
	window->view = NULL;
	window->reload = FALSE;
	window->back_history = NULL;
	window->forward_history = NULL;
	window->nohistory = FALSE;
	window->title_prefix = g_strdup(APPNAME);
	
	char *cwd = getcwd(NULL, 0);
	gchar *uri = g_strdup_printf("file://%s/", cwd);
	window->baseURI = str_to_url (uri);
	window->home_url = g_strdup(uri);
	g_print("Initial baseURI: %s\n", uri);
	g_free(uri);
	free(cwd);

	window->entry = gtk_entry_new ();
	g_signal_connect (window->entry, "activate",
			  G_CALLBACK (browser_window_entry_activate), window);

	g_signal_connect (G_OBJECT(window), "configure-event",
			  G_CALLBACK (browser_window_reload_delayed), window);

	back_button = GTK_WIDGET(gtk_tool_button_new_from_stock(GTK_STOCK_GO_BACK));
	g_signal_connect (back_button, "clicked",
			  G_CALLBACK (browser_window_back_clicked), window);

	forward_button = GTK_WIDGET(gtk_tool_button_new_from_stock(GTK_STOCK_GO_FORWARD));
	g_signal_connect (forward_button, "clicked",
			  G_CALLBACK (browser_window_forward_clicked), window);	

	refresh_button = GTK_WIDGET(gtk_tool_button_new_from_stock(GTK_STOCK_REFRESH));
	g_signal_connect_swapped (refresh_button, "clicked",
			  G_CALLBACK (browser_window_reload), window);	

	home_button = GTK_WIDGET(gtk_tool_button_new_from_stock(GTK_STOCK_HOME));
	g_signal_connect_swapped (home_button, "clicked",
			  G_CALLBACK (browser_window_load_home), window);	

	stop_button = GTK_WIDGET(gtk_tool_button_new_from_stock(GTK_STOCK_STOP));
	g_signal_connect_swapped (stop_button, "clicked",
			  G_CALLBACK (browser_window_stop_loading), window);	
	
	spinner = gtk_spin_button_new_with_range (1, 999, 1);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (spinner), 100.0);
	g_signal_connect (G_OBJECT (spinner), "value_changed",
				  G_CALLBACK (zoom_changed), window);	
	
	accels = gtk_accel_group_new();
	window->item_factory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<main>", accels);
	gtk_item_factory_create_items (window->item_factory, G_N_ELEMENTS (menu_items), menu_items, window);
	gtk_window_add_accel_group (GTK_WINDOW(window), accels);

	window->vbox = gtk_vbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER (window), window->vbox);
	
	gtk_box_pack_start (GTK_BOX (window->vbox),
			    gtk_item_factory_get_widget (window->item_factory, "<main>"),
			    FALSE, FALSE, 0);

	hbox = gtk_hbox_new (FALSE, 0);

	gtk_box_pack_start (GTK_BOX (hbox), back_button, FALSE, FALSE, 0);				    
	gtk_box_pack_start (GTK_BOX (hbox),forward_button, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (hbox),home_button, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (hbox),refresh_button, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (hbox),stop_button, FALSE, FALSE, 0);
			    
	gtk_box_pack_start (GTK_BOX (hbox),
			    gtk_label_new ("URL: "),
			    FALSE, FALSE, 0);
			    
	gtk_box_pack_start (GTK_BOX (hbox),
			    window->entry, TRUE, TRUE, 0);
				  
	gtk_box_pack_start (GTK_BOX (hbox),
			    spinner,
			    FALSE, FALSE, 0);

	frame = gtk_frame_new (NULL);
	gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_OUT);
	gtk_container_border_width (GTK_CONTAINER (hbox), 3);
	gtk_container_add (GTK_CONTAINER (frame), hbox);

	gtk_box_pack_start (GTK_BOX (window->vbox),
		frame, FALSE, FALSE, 0);

	window->status_bar = gtk_statusbar_new ();
	gtk_widget_show (window->status_bar);

	gtk_box_pack_end (GTK_BOX (window->vbox),
			    window->status_bar, FALSE, FALSE, 0);

	gtk_window_set_title (GTK_WINDOW (window), window->title_prefix);
	gtk_window_set_default_size (GTK_WINDOW (window), 500, 400);

	gtk_window_set_icon(GTK_WINDOW (window),
		gtk_widget_render_icon (GTK_WIDGET(window),
						  GTK_STOCK_ABOUT,
						  GTK_ICON_SIZE_MENU,
						  "BrowserWindow"));

	gtk_widget_show_all (window->vbox);
	gtk_widget_grab_focus (window->entry);	
}

// ======== browser window gobject class =========
GtkType
browser_window_get_type (void)
{
  static GtkType window_type = 0;

  if (!window_type)
    {
	    static const GTypeInfo window_info =
	    {
		    sizeof (BrowserWindowClass),
		    NULL,
		    NULL,
		    (GClassInitFunc) browser_window_class_init,
		    NULL,
		    NULL,
		    sizeof (BrowserWindow),
		    16,
		    (GtkObjectInitFunc) browser_window_init,
	    };
	    
	    g_print ("gtk_window: %d browser_window: %d\n", sizeof (GtkWindowClass), sizeof (BrowserWindowClass));
	    

	    window_type = g_type_register_static (GTK_TYPE_WINDOW, "BrowserWindow", &window_info, 0);
    }

  return window_type;
}

static void
browser_window_instance_finalize (GObject *obj)
{
	BrowserWindow *window = (BrowserWindow *) obj;
	BROWSER_IS_WINDOW(window);
	
	// clear out private data
	url_free (window->baseURI); window->baseURI = NULL;
	g_free (window->title_prefix); window->title_prefix = NULL;
	history_clear (&window->back_history); 
	history_clear (&window->forward_history);
	g_free (window->home_url);
	g_source_remove (window->stale_connection_remover);
	
	// chain to parent
	G_OBJECT_CLASS(parent_class)->finalize(obj);
}

static void
browser_window_class_init (BrowserWindowClass *klass)
{
	GtkObjectClass *object_class;

	object_class = (GtkObjectClass*) klass;
	G_OBJECT_CLASS(object_class)->finalize = browser_window_instance_finalize;

	parent_class = gtk_type_class (GTK_TYPE_WINDOW);
}


GtkWidget *
browser_window_new (HtmlDocument *doc)
{
	BrowserWindow *window;

	window = g_object_new (BROWSER_TYPE_WINDOW, NULL);

	if (!doc) {
		window->doc = html_document_new ();

		g_signal_connect (G_OBJECT (window->doc), "link_clicked",
				  G_CALLBACK (link_clicked), window);

		g_signal_connect (G_OBJECT (window->doc), "title_changed",
				  G_CALLBACK (title_changed), window);

		g_signal_connect (G_OBJECT (window->doc), "request_url",
				  GTK_SIGNAL_FUNC (url_requested), window);

		g_signal_connect (G_OBJECT (window->doc), "submit",
				  GTK_SIGNAL_FUNC (on_submit), window);
	}
	else 
		window->doc = doc;

	browser_window_set_view (window, HTML_VIEW (html_view_new ()));
	html_view_set_document (window->view, window->doc);
	window->stale_connection_remover = 
		g_timeout_add(CONNECTION_CLEANER_PERIOD,
			on_idle_remove_stale_connections, window);

	return GTK_WIDGET (window);
}

// === public APIs
void 
browser_window_set_debug(GtkWidget *widget, gboolean enable)
{
	BrowserWindow *window = BROWSER_WINDOW(widget);	
	g_assert (BROWSER_IS_WINDOW(window));

	window->debug_connection = enable;
}

void 
browser_window_set_insecure(GtkWidget *widget, gboolean enable)
{
	BrowserWindow *window = BROWSER_WINDOW(widget);
	g_assert (BROWSER_IS_WINDOW(window));

	window->allow_insecure = enable;
}

void 
browser_window_set_online(GtkWidget *widget, gboolean enable)
{	
	BrowserWindow *window = BROWSER_WINDOW(widget);
	g_assert (BROWSER_IS_WINDOW(window));

	window->allow_online = enable;
	
	// Set the initial state of the Online menu item
	GtkWidget *online = gtk_item_factory_get_item(window->item_factory, "/File/Online");
	if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(online)) != enable) {
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(online), enable);
	}
}

void 
browser_window_set_title(GtkWidget *widget, gchar *title)
{
	BrowserWindow *window = BROWSER_WINDOW(widget);
	g_assert (BROWSER_IS_WINDOW(window));

	if (window->title_prefix) g_free(window->title_prefix);
	window->title_prefix = g_strdup(title);
	gtk_window_set_title (GTK_WINDOW(window), window->title_prefix);
}

void 
browser_window_load_url(GtkWidget *widget, const gchar *url)
{
	BrowserWindow *window = BROWSER_WINDOW(widget);
	g_assert (BROWSER_IS_WINDOW(window));

	window->reload = TRUE;
	//link_clicked (window->doc, url, window);
	RedirectContext *ctx = g_new0(RedirectContext, 1);
	ctx->window = window;
	ctx->url = g_strdup(url);
	g_idle_add (on_idle_redirect, ctx);
}

void
browser_window_load_file (GtkWidget *widget, const gchar *name)
{
	BrowserWindow *window = BROWSER_WINDOW(widget);

	name = realpath(name, NULL);
	gchar *copy = g_strdup(name);
	gchar *dir = g_strdup_printf("file://%s/", dirname(copy));
	//g_print("%s\n", dir); 
	
	set_base(window, dir);
	load_file (window->doc, name, NULL);

	g_free(copy); 
	g_free(dir);
	free((void *)name);
}
void
browser_window_set_home_url (GtkWidget *widget, const gchar *url)
{
	BrowserWindow *window = BROWSER_WINDOW(widget);	
	g_assert (BROWSER_IS_WINDOW(window));
	
	if (window->home_url) g_free(window->home_url);
	window->home_url = g_strdup(url);
}
