#include <gtk/gtk.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "url-utils.h"

//============== generic helpers =============
gchar *dequote(gchar *instr) {
	if (!instr) return NULL;
	// don't dequote, because it breaks loading files with ' " \ in their paths
	return g_strdup(instr);
	// a silly state-machine to "parse" quoting
	#define NQ  0
	#define SQ  1
	#define DQ  2
	int next[3][4] = { // state, input-token
		{NQ, SQ, DQ},  // no-quote with input of no-quote, single, double-quotes
		{SQ, NQ, SQ},  // single-quote with input of no-quote, single, double-quotes
		{DQ, DQ, NQ}   // double-quote with input of no-quote, single, double-quotes
	};
	int copy[3][3] = { // state, input
		{1, 0, 0},     // no-quote with input of no-quote, single, double-quotes
		{1, 0, 1},     // single-quote with input of no-quote, single, double-quotes
		{1, 1, 0}      // double-quote with input of no-quote, single, double-quotes
	};
	int literal_allowed[3] = { // state
		1, 0, 1  // quote only allowed during no-quote and double-quotes
	};
	int state; // as #defined above: 0=no quote, 1=in-single-quote, 2=in-double-quote
	int literal;
	int tok;
	
	gchar *outstr = g_malloc0(strlen(instr)+1);
	gchar *inp, *outp, c;
	
	state = literal = 0;
	for (inp=instr, outp=outstr; *inp; inp++) {
		c = *inp;
		tok = NQ;
		if (!literal) {
			switch (c) {
				case '\'' : tok=SQ; break;
				case '\"' : tok=DQ; break;
				case '\\' : literal = literal_allowed[state]; 
				            if (literal) continue;
			}
		} else {
			// supported literals
			switch (*inp) {
				case 't' : c = '\t'; break;
				case 'n' : c = '\n'; break;
				case 'r' : c = '\r'; break;
				case 'b' : c = '\b'; break;
				case 'a' : c = '\a'; break;
				case '[' : c = '\['; break;
			}
			literal=0; 
		}
		if (copy[state][tok]) *(outp++) = c;
		state = next[state][tok];
	}
	//g_print("dequote done\n");
	return outstr;
}

gchar *url_make_canonical(const gchar* path) {
	if (!path) return NULL;
	
	gchar *ret;
	gchar **strv = g_strsplit(path,"/",0);
	int elements = g_strv_length(strv);
	gchar **strv2 = g_malloc0(sizeof(void *) * (elements+1));
	int i,j;
	for (i=j=0; i<elements; i++) {
		if (!*(strv[i]) && i>0 && i<elements-1) {
			// empty component - skip, but not the first or the last
			// (leading or trailing slash)
			continue;
		} else if (strcmp(strv[i],".")==0) {
			// single dot, treat as empty component
			continue;			
		} else if (strcmp(strv[i],"..")==0) {
			// double-dot, go up if possible
			if (j>0) j--;					
			if (strv2[j]) g_free(strv2[j]);
			strv2[j]=NULL;
		} else {
			// copy path
			strv2[j++] = g_strdup(strv[i]);
		}
	}
	gchar *tmp = g_strjoinv("/", strv2);
	ret = dequote(tmp);
	g_free(tmp);
	g_strfreev(strv2);
	g_strfreev(strv);
	return ret;
}

void add_trailing_slash(URL *url) {
	if (!url) return;
	
	gchar *tmp = url->path;
	struct stat sb;	
	if (!url->proto || strcmp(url->proto,"file")==0) 
	{
		if (stat(tmp, &sb)==0 && 
			S_ISDIR(sb.st_mode) && 
			(tmp[strlen(tmp)-1] != '/'))
		{
			url->path = g_strconcat(tmp, "/", NULL);
			g_free(tmp);
			//g_print("final %s\n", url->path);
		}				
	}
}

//============== url helpers =============
URL* url_new() {
	return g_malloc0(sizeof(URL));
}

void url_free(URL *u) {
	if (!u) return;

	g_free(u->proto);
	g_free(u->host);
	g_free(u->user);
	g_free(u->pass);
	g_free(u->path);
	g_free(u->query);
	g_free(u->anchor);
	g_free(u);
}

URL* url_copy(URL *u) {
	if (!u) return NULL;
		
	URL *ret = url_new();
	ret->proto = g_strdup(u->proto);
	ret->host  = g_strdup(u->host);
	ret->user  = g_strdup(u->user);
	ret->pass  = g_strdup(u->pass);
	ret->path  = g_strdup(u->path);
	ret->query = g_strdup(u->query);
	ret->anchor = g_strdup(u->anchor);
	return ret;
}


// ========= actual url routines ========== 
gchar *url_to_str(URL* url) {
	if (!url) return NULL;
	
	gchar *ret = NULL;
	GString *s = g_string_sized_new(256);
	if (url->proto)
		g_string_append_printf(s,"%s://",url->proto);
	if (url->host) {
		if (url->pass) {
			g_string_append_printf(s,"%s:%s@%s",url->user,url->pass,url->host);	
		} else if (url->user) {
			g_string_append_printf(s,"%s@%s",url->user,url->host);	
		} else {
			g_string_append_printf(s,"%s",url->host);	
		}
	}
	if (url->path)
		g_string_append(s, url->path);
	if (url->query)
		g_string_append_printf(s, "?%s",url->query);	
	if (url->anchor)
		g_string_append_printf(s, "#%s",url->anchor);	
	
	ret = s->str;
	g_string_free(s,FALSE);
	return ret;
}

// url parser
URL* str_to_url(const gchar *url) {
	if (!url) return NULL;
	
	URL *ret = url_new();
	gchar *copy, *start, *p, *tmp;

	copy = g_malloc0(strlen(url)+1+1); // extra for null and guard 
	strcpy(copy, url);
	start = copy;
	
	// if there any proto?	
	if (!(*start=='?' || *start=='#' || *start=='/')) {
		p = strchr(start, ':');
		if (p) {
			*p=0;
			ret->proto = g_strdup(start);
			start = p+1;
		}
	}

	// if there any host?
	if (*start && *start=='/' && *(start+1)=='/')
	{
		p = (start+=2);
		while (*p && *p!='/' && *p!='#' && *p!='?') p++;
		*p=0;
		tmp = start;		
		start = p+1;
		
		// extract user, pass, host
		p = strchr(tmp,'@');
		if (p) {
			*p = 0;
			ret->host = g_strdup(p+1);
			p = strchr(tmp, ':');
			if (p) {
				*p = 0;
				ret->user = g_strdup(tmp);
				ret->pass = g_strdup(p+1);
			} else {
				ret->user = g_strdup(tmp);
			}
		} else {
			// host only
			ret->host = g_strdup(tmp);
		}
	} 
	
	// check for anchor/query first (in that order)
	p = strchr(start, '#'); 
	if (p) {
		*p=0;
		ret->anchor = g_strdup(p+1);
	}
	p = strchr(start, '?'); 
	if (p) {
		*p=0;
		ret->query = g_strdup(p+1);
	}

	// the rest is path, if there is anything left
	if (*start) {
		tmp = g_strconcat("/", start, NULL);
		ret->path = url_make_canonical(tmp);
		g_free(tmp);
	}
	
	g_free(copy);
	add_trailing_slash (ret);
	return ret;
}


// --- get components 
gchar *url_get_proto(const gchar *url) {
	if (!url) return NULL;
	URL *u = str_to_url (url);
	gchar *ret = g_strdup(u->proto);
	url_free(u);
	return ret;	
}

gchar *url_get_host(const gchar *url) {
	if (!url) return NULL;
	URL *u = str_to_url (url);
	gchar *ret = g_strdup(u->host);
	url_free(u);
	return ret;		
}

gchar *url_get_user(const gchar *url) {
	if (!url) return NULL;
	URL *u = str_to_url (url);
	gchar *ret = g_strdup(u->user);
	url_free(u);
	return ret;		
}

gchar *url_get_pass(const gchar *url) {
	if (!url) return NULL;
	URL *u = str_to_url (url);
	gchar *ret = g_strdup(u->pass);
	url_free(u);
	return ret;			
}

gchar *url_get_path(const gchar * url) {
	if (!url) return NULL;
	URL *u = str_to_url (url);
	gchar *ret = g_strdup(u->path);
	url_free(u);
	return ret;	
}

gchar *url_get_query(const gchar *url) {
	if (!url) return NULL;
	URL *u = str_to_url (url);
	gchar *ret = g_strdup(u->query);
	url_free(u);
	return ret;	
}

gchar *url_get_anchor(const gchar *url) {
	if (!url) return NULL;
	URL *u = str_to_url (url);
	gchar *ret = g_strdup(u->anchor);
	url_free(u);
	return ret;	
}


// out: NULL = can't process, otherwise new URL is given.
URL* url_set_relative (URL *base, const gchar *new_url) {	
	if (!new_url) 
		return url_copy(base);
	
	if (!base) {
		return str_to_url(new_url);
	}
	
	// both exist, so process now
	gchar *tmp;
	URL *ret = NULL;
	URL *b = base;
	URL *u = str_to_url(new_url);		
		
	if (u->proto) {
		// if there is proto, it is absolute
		return u;
	} else {		
		// relative, get how relative it is.
		ret = g_malloc0(sizeof(URL));
		ret->proto = g_strdup(b->proto);
		if (u->host) {
			// host, user, name, path, anchor, query
			ret->host = g_strdup(u->host);
			ret->user = g_strdup(u->user);
			ret->pass = g_strdup(u->pass);
			ret->path = g_strdup(u->path); 
			ret->query = g_strdup(u->query);
			ret->anchor = g_strdup(u->anchor);
			
		} else if (u->path) {		
			// path (absolute/relative)
			ret->host = g_strdup(b->host);
			ret->user = g_strdup(b->user);
			ret->pass = g_strdup(b->pass);
			ret->query = g_strdup(u->query);
			ret->anchor = g_strdup(u->anchor);
			
			gchar *copy = g_strdup(new_url);
			gchar *tmp;
			if (strchr(copy,'?')) *(strchr(copy,'?'))=0;
			if (strchr(copy,'#')) *(strchr(copy,'#'))=0;
			
			if (!b->path || b->path[0]==0 || copy[0]=='/') {
				// absolute
				tmp = g_strconcat("/", copy, NULL);
			} else {
				// relative
				int n = strlen(b->path);
				if (b->path[n-1]=='/') tmp = g_strconcat(b->path, copy, NULL);
				else tmp = g_strconcat(b->path, "/../", copy, NULL);
			}
			ret->path = url_make_canonical(tmp);
			g_free(tmp);						
			g_free(copy);
			add_trailing_slash(ret);
		} else if (u->query) {
			// query, anchor
			ret->host = g_strdup(b->host);
			ret->user = g_strdup(b->user);
			ret->pass = g_strdup(b->pass);
			ret->path = g_strdup(b->path);
			ret->query = g_strdup(u->query);
			ret->anchor = g_strdup(u->anchor);
		} else if (u->anchor) {
			// anchor
			ret->host = g_strdup(b->host);
			ret->user = g_strdup(b->user);
			ret->pass = g_strdup(b->pass);
			ret->path = g_strdup(b->path);
			ret->query = g_strdup(b->query);
			ret->anchor = g_strdup(u->anchor);
		}
		url_free(u);
		return ret;
	}
}
